@@ -72,6 +72,19 @@ unsigned int ExecedProcessCacher::cache_format_ = 0;
 /* singleton*/
 ExecedProcessCacher* execed_process_cacher;
 
+static void write_cache_format_file(const char* cache_format_file) {
+  FILE* f;
+  if (!(f = fopen(cache_format_file, "w"))) {
+    fb_perror("creating cache-format file failed");
+    exit(EXIT_FAILURE);
+  }
+  if (fprintf(f, "%d\n", kCacheFormatVersion) <= 0) {
+    fb_perror("writing cache-format file failed");
+    exit(EXIT_FAILURE);
+  }
+  fclose(f);
+}
+
 void ExecedProcessCacher::init(const libconfig::Config* cfg) {
   std::string cache_dir;
   char* cache_dir_env;
@@ -130,20 +143,22 @@ void ExecedProcessCacher::init(const libconfig::Config* cfg) {
       } else {
         /* Cache is in a prior format. Either use it considering the differences where needed
          * or upgrade it. */
+        /* For now the cache is just cleared. */
+        fb_info("Cache format version is outdated, clearing the cache");
+        // remove all files in the cache dir, but keep the dir itself and the cache-format file
+        for (const auto& entry : std::filesystem::directory_iterator(cache_dir)) {
+          std::error_code ec;
+          std::filesystem::remove_all(entry.path(), ec);
+          if (ec) {
+            fb_error("Failed to clear cache: " + ec.message());
+          }
+        }
+        write_cache_format_file(cache_format_file);
       }
       fclose(f);
     }
   } else {
-    FILE* f;
-    if (!(f = fopen(cache_format_file, "wx"))) {
-      fb_perror("creating cache-format file failed");
-      exit(EXIT_FAILURE);
-    }
-    if (fprintf(f, "%d\n", kCacheFormatVersion) <= 0) {
-      fb_perror("writing cache-format file failed");
-      exit(EXIT_FAILURE);
-    }
-    fclose(f);
+    write_cache_format_file(cache_format_file);
   }
   free(cache_format_file);
 
@@ -1976,6 +1991,10 @@ void ExecedProcessCacher::reset_stored_stats() {
 }
 
 void ExecedProcessCacher::update_stored_stats() {
+  if (no_store_) {
+    /* In read-only mode, don't update cache metadata */
+    return;
+  }
   // FIXME(rbalint) There is a slight chance for two parallel builds updating the stats at the
   // same time making them inaccurate.
   add_stored_stats();
@@ -2014,6 +2033,10 @@ void ExecedProcessCacher::read_stored_cached_bytes() {
 }
 
 void ExecedProcessCacher::update_stored_bytes() {
+  if (no_store_) {
+    /* In read-only mode, don't update cache metadata */
+    return;
+  }
   // FIXME(rbalint) There is a slight chance for two parallel builds updating the size at the
   // same time making the file content inaccurate.
   const std::string size_file = cache_dir_ + "/" + kCacheSizeFile;
@@ -2042,6 +2065,10 @@ bool ExecedProcessCacher::is_gc_needed() const {
 }
 
 void ExecedProcessCacher::gc() {
+  if (no_fetch_ || no_store_) {
+    fb_info("Garbage collection is disabled in read-only and no-fetch mode, skipping.");
+    return;
+  }
   gc_runs_++;
   /* Remove unusable entries first. */
   tsl::hopscotch_set<AsciiHash> referenced_blobs {};
